function [out, var] = SFFS_reg(tbl_train, label_train, sett) 

% [out, var] = SFFS_reg(tbl_train, label_train, sett)
%
% This function performs sequential floating forward selection using the
% input variables: (tbl_train label_train). The columns of tbl_train does
% correspond to the features and the rows corresponds to the observations
%
% Starting from an empty feature set one feature at a time is added based
% on which feature gives the best estimation (reg) performance, when used
% together with the previously selected features. The implementation does
% the cross-validation process to select the best feature set using input
% settings of the regression algorithm, scroring function, etc.
%
% In this implementation the mean absolute error is used to decide whether 
% to include the feature in the selected feature set. Tolerance parameter
% t specifies how many features are allowed to be added without improving
% upon the best regression score obtained during the search. The maximum
% acceptable size of feature set can be set by parameter N. Increasing the
% size of the feature set stops when either of the conditions (related to
% t and N) is reached.
%
%       ------------------------------------------------------------------
%       Required inputs:
% 
%           tbl_train:          Input feature matrix. Columns: features; 
%                               rows: observations. The matrix is numeric
%                               and it represent the training and testing
%                               set of data for SFFS process.
%
%           label_train:        Input labels that corresponds to classes
%                               of the observations from tbl_train. This
%                               algorithm requires labels to be numeric,
%                               where: classes are continuous numbers
%
%       ------------------------------------------------------------------
%       Optional inputs:
%
%           num_cross_val.:     The number of cross validation runs that 
%                               the algorithm perform in every search for 
%                               the best features The score is consequen.
%                               averaged in order to obtain better estim.  
%                               of the model accuracy.
%                               Default: 100 cross-validation runs
%
%           k_fold:             The number of cross-validation folds in 
%                               every cross-validation run. 
%                               Default: 10-fold cross-validation
%
%           classifier:         The classifier to use in oreder to do the
%                               estimation (reg) of the testing data. This 
%                               is specified in a structure holding assoc.
%                               member variables settings for the cls
%                               Default: .alg   - 'cart'
%                                        .no settings so far (TODO)
%
%           score:              The scoring algorithm that is used to find
%                               the features tobe included or excluded in 
%                               / from the best feature set.
%                               Default: 'mae' - mean absolute error
%
%           t:                  The number of features that is allowed to
%                               be included into the selected feature set
%                               without improving obtained best overall 
%                               estimation (reg) score previously seen 
%                               with smaller feature sets.
%                               Default: 3.
%
%           N:                  The maximum number of features, that is to 
%                               be selected as the best possible feat. set
%                               Default: size(tbl_train, 2)
%
%       ------------------------------------------------------------------
%       Outputs:
%
%           out:                Structure holding the output variables:
%                               out.S             - feature indices
%                               out.SCORE         - regression score
%                               out.RMSE          - Root mean sqr. error
%                               out.RHO           - Spearman's corr. c. 
%                               out.P             - rho significance lvl. 
%
%           var:                Structure, that holds the change of output
%                               variables during the function runtime. The
%                               score, acc, sen etc. are stored in the var
%                               VAR.score         - regression score
%                               VAR.rmse          - root mean squared err.
%                               VAR.rho           - Spearman's corr. coeff
%                               VAR.p             - significance level 
%                               VAR.num_features  - num. of features in S
%
% Implemented according to:
% A. W. Whitney: "A direct method of nonparametric measurement selection",
% IEEE Transactions on Computers, vol. C-20, no. 9, pp. 1100-1103,
% September 1971.
%
%
%
% --
% ing. Zoltn Gal
% xgalaz00@stud.feec.vutbr.cz       
% 
% Department of Telecommunications
% Faculty of Electrical Engineering and Communication
% Brno University of Technology

%% Define the optional variable settings
%  [A] Define all the possible estimation (reg) algorithms and also its
%      associated member variables. The supported regression algorithms 
%      are listed bellow with the description of settings + its default 
%      values selected if the option of the classifier is not specified, 
%      or is specified incorrectly (corrections are displayed into the 
%      console)

%      Regression algorithms and settings:
%           a) Classification and Regression Trees algorithm
%              no settings so far (TODO)

classifier.def_alg  = 'cart';
classifier.pos_alg  = 'cart';
classifier.cart.alg = 'cart';

%  [B] Define the regression scoring algorithm supported. The list of all
%      algorithms is listed bellow. Default scoring algorithm is used if
%      input structure does not contain information about algorithm to 
%      use or if it is incorrectly inserted. Default algorithm: 'mae'
%
%      Scoring algorithms (used in regression (estimation) process):
%           - ae    (absolute error)     
%           - mae   (mean absolute error)
%           - se    (squared error)
%           - sle   (squared log error)
%           - mse   (mean squared error)
%           - msle  (mean squared log error)
%           - rmse  (root mean squared error)
%           - rmsle (root mean squared log error)
%           - gii   (gini index)
%           - ngii  (normalized gini index)

score.pos_alg = {{'ae'    '(absolute error)'};               ...
                 {'mae'   '(mean absolute error)'};          ...
                 {'se'    '(squared error)'};                ...
                 {'sle'   '(squared log error)'};            ...
                 {'mse'   '(mean squared error)'};           ...
                 {'msle'  '(mean squared log error)'};       ...
                 {'rmse'  '(root mean squared error)'};      ...
                 {'rmsle' '(root mean squared log error)'};  ...
                 {'gii'   '(gini index)'};                   ...
                 {'ngii'  '(normalized gini index)'}};
score.def_alg =  'mae';

%% Set the default/helper variables
%  [A] Helper variables that allow optional variables check and a console 
%      output. Those variables contain the helper variables (set / unset)
%      and default variables that hold the default values for unspecified
%      variables (empty, not specified or out of range). The algorithm is
%      implemented to apply the predefined value

DEF_NUM_CROSS_VAL  = true; % Helpser var.: number of C.V. rounds specified
DEF_K_FOLD         = true; % Helpser var.: k (number) C.V. folds specified
DEF_CLASSIFIER     = true; % Helpser var.: classifier algorithm specified
DEF_SCORE          = true; % Helpser var.: cls scoring algorithm specified
DEF_MAX_NUM_FEAT   = true; % Helpser var.: max. num. of feat. is specified
DEF_ALLOWED_FEAT   = true; % Helpser var.: num. look ahead feat. specified 
                                    
%  [B] Default settings for the SFFS algorithm. It includes the setting of
%      the number of cross-validation runs and k-folds of cross-validation
%      then the classifier algorithm, scoring function, number of features
%      to look ahead without closing the algorithm and the maximum number
%      of features to be inside the selected feature set

DEFAULT_NUM_CROSS_VAL = 100;                % DEFAULT: num. of C.V. rounds
DEFAULT_K_FOLD        = 10;                 % DEFAULT: nun. of C.V. folds
DEFAULT_CLASSIFIER    = classifier.def_alg; % DEFAULT: classif. algorithm
DEFAULT_SCORE         = score.def_alg;      % DEFAULT: scoring algorithm
DEFAULT_ALLOWED_FEAT  = 3;                  % DEFAULT: look ahead features
DEFAULT_MAX_NUM_FEAT  = ...                 % DEFAULT: max. num. features
            size(tbl_train, 2);             %          (number of features)

%% Paths and variables
if ((nargin < 3) || (isempty(sett)))  
    
    % Set the number of cross-validations
    sett.num_cross_validations = DEFAULT_NUM_CROSS_VAL;
    DEF_NUM_CROSS_VAL = false;
    
    % Set the k-folds of cross-validation 
    sett.k_fold = DEFAULT_K_FOLD;
    DEF_K_FOLD = false;
    
    % Set the classification algorithm
    sett.classifier = struct();
    sett.classifier.alg = DEFAULT_CLASSIFIER;
    DEF_CLASSIFIER  = false;
    
    % Set the scoring function of classification
    sett.score = DEFAULT_SCORE;
    DEF_SCORE = false;
    
    % Set the maximum number of added (ahead) features
    sett.t = DEFAULT_ALLOWED_FEAT;
    DEF_ALLOWED_FEAT = false;
    
    % Set the maximum number of selected features
    sett.N = DEFAULT_MAX_NUM_FEAT;
    DEF_MAX_NUM_FEAT = false;
    
else
    
    % Set the number of cross-validations
    if (~isfield(sett, 'num_cross_validations'))
        sett.num_cross_validations = DEFAULT_NUM_CROSS_VAL;
        DEF_NUM_CROSS_VAL = false; 
    end
    
    % Set the k-folds of cross-validation 
    if (~isfield(sett, 'k_fold'))
        sett.k_fold = DEFAULT_K_FOLD;
        DEF_K_FOLD = false;
    end
    
    % Set the classification algorithm 
    if (~isfield(sett, 'classifier'))
        sett.classifier = struct();
        sett.classifier.alg = DEFAULT_CLASSIFIER;
        DEF_CLASSIFIER  = false;
    end
    
    % Set the scoring function of classification
    if (~isfield(sett, 'score'))
        sett.score = DEFAULT_SCORE;
        DEF_SCORE = false;
    end
    
    % Set the maximum number of added (ahead) features
    if (~isfield(sett, 't'))
        sett.t = DEFAULT_ALLOWED_FEAT;
        DEF_ALLOWED_FEAT = false;
    end
    
    % Set the maximum number of selected features
    if (~isfield(sett, 'N'))
        sett.N = DEFAULT_MAX_NUM_FEAT;
        DEF_MAX_NUM_FEAT = false;
    end
end

sett.score = lower(sett.score);
sett.classifier.alg = lower(sett.classifier.alg);

%% Test the input data   -
%  [A] Check if the number of observations in 'tbl_train' is equal to the
%      number of observations in 'label_train'. The algorithm demands the 
%      same number of observations (rows) = e.g. patients (PD/HC, etc.)

if (size(tbl_train, 1) ~= size(label_train, 1))
    error('The number of observations in tbl_train ~= label_train');
end

%  [B] Optional variables handling routines to prepare the console output
%
%      Handle: The number of cross-validation runs
%              DEFAULT: 100 cross-validation runs
if (~DEF_NUM_CROSS_VAL)
    STR_NUM_CROSS_VAL = num2str(sett.num_cross_validations);
    STR_NUM_CROSS_VAL_STATE = ' (DEFAULT)';
elseif ((DEF_NUM_CROSS_VAL) && (sett.num_cross_validations <= 0))
    sett.num_cross_validations = DEFAULT_NUM_CROSS_VAL;
    STR_NUM_CROSS_VAL = num2str(sett.num_cross_validations);
    STR_NUM_CROSS_VAL_STATE = ' (AUTOMATIC)';
else
    STR_NUM_CROSS_VAL = num2str(sett.num_cross_validations);
    STR_NUM_CROSS_VAL_STATE = '';
end

%      Handle: The number of cross-validation folds (k)
%              DEFAULT: 10 cross-validation folds
if (~DEF_K_FOLD)
    STR_K_FOLD = [num2str(sett.k_fold) '-fold'];
    STR_K_FOLD_STATE = ' (DEFAULT)';
elseif ((DEF_K_FOLD) && ((sett.k_fold <= 0) || ...
        (sett.k_fold > floor(size(tbl_train, 1)/2))))
    sett.k_fold = DEFAULT_K_FOLD;
    STR_K_FOLD = [num2str(sett.k_fold) '-fold'];
    STR_K_FOLD_STATE = ' (AUTOMATIC)';
elseif ((DEF_K_FOLD) && (sett.k_fold == 1))    
    STR_K_FOLD = 'L-O-OUT';
    STR_K_FOLD_STATE = '';

    if (sett.num_cross_validations ~= 1)
        sett.num_cross_validations = 1;
        STR_NUM_CROSS_VAL = num2str(sett.num_cross_validations);
        STR_NUM_CROSS_VAL_STATE = ' (AUTOMATIC)';
    end
else       
    STR_K_FOLD = [num2str(sett.k_fold) '-fold'];
    STR_K_FOLD_STATE = '';
end

%      Handle: Regression (estim) algorithm used in the feature selection
%              DEFAULT: Cart: Classification and Regression Trees 
if (~DEF_CLASSIFIER)
    STR_CLASSIFIER = sett.classifier.alg;
    STR_CLASSIFIER_STATE = ' (DEFAULT)';
    sett.classifier = ...
            check_regression_sett(sett.classifier, classifier);
else

    % check the classifier (set the regression algorithm)
    sett.classifier.alg = check_regression_alg(sett.classifier.alg);
    if (~isempty(sett.classifier.alg))
         STR_CLASSIFIER = sett.classifier.alg;
         STR_CLASSIFIER_STATE = '';
         sett.classifier = ...
            check_regression_sett(sett.classifier, classifier);
    else
        disp(' '                                               );
        disp('Unknown classifier, set to DEFAULT: cart'        );
        disp('Known classifiers:'                              );
        disp('   - cart  (Classification and Regression Trees)');
        
        STR_CLASSIFIER = DEFAULT_CLASSIFIER;
        STR_CLASSIFIER_STATE = ' (AUTOMATIC)';
        sett.classifier = ...
            check_regression_sett(sett.classifier, classifier);
    end
end

%      Handle: The ccoring method used used in the feature selection
%              DEFAULT: 'mae' (mean absolute error)
if (~DEF_SCORE)
    STR_SCORE = sett.score;
    STR_SCORE_STATE = ' (DEFAULT)';
else

    % check the regression score (set the regression score)
    sett.score = check_regression_score(sett.score);
    if (~isempty(sett.score))
         STR_SCORE = sett.score;
         STR_SCORE_STATE = '';
    else
        
        disp(' '                                           );
        disp('Unknown scoring method, set to DEFAULT: mae' );
        disp('Known scoring methods:'                      );
        disp('   - ae    (absolute error)'                 );            
        disp('   - mae   (mean absolute error)'            );
        disp('   - se    (squared error)'                  );
        disp('   - sle   (squared log error)'              );
        disp('   - mse   (mean squared error)'             );
        disp('   - msle  (mean squared log error)'         );
        disp('   - rmse  (root mean squared error)'        );
        disp('   - rmsle (root mean squared log error)'    );
        disp('   - gii   (gini index)'                     );
        disp('   - ngii  (normalized gini index)'          );

        STR_SCORE = DEFAULT_SCORE;
        STR_SCORE_STATE = ' (AUTOMATIC)';
    end
end

%      Handle: Maximum number of features to look ahead
%              BOUNDS:  <0, floor(size(tbl_train, 2))/2)  Z 
%              DEFAULT: 3 allowed features (local max. possibility)
if (~DEF_ALLOWED_FEAT)
    STR_ALLOWED_FEAT = num2str(sett.t);
    STR_ALLOWED_FEAT_STATE = ' (DEFAULT)';
else
    if (sett.t >= floor(size(tbl_train, 2)/2) || (sett.t < 0))
        disp(' '                                                   );
        disp('Improper number of added features, set to DEFAULT: 3');
        disp('Allowed values <0, floor(size(tbl_train, 2))/2)'     );
        disp('WARNING: values  Z'                                 );

        sett.t = DEFAULT_ALLOWED_FEAT;
        STR_ALLOWED_FEAT = num2str(sett.t);
        STR_ALLOWED_FEAT_STATE = ' (AUTOMATIC)';
    else
        STR_ALLOWED_FEAT = num2str(sett.t);
        STR_ALLOWED_FEAT_STATE = '';
    end
end

%      Handle: Maximum number of features selected into output feat. set
%              BOUNDS:  <1, size(tbl_train, 2))  Z 
%              DEFAULT: max. num. of features == the number of features
if (~DEF_MAX_NUM_FEAT)
    STR_MAX_NUM_FEAT = num2str(sett.N);
    STR_MAX_NUM_FEAT_STATE = ' (DEFAULT)';
else
    if (sett.N >= size(tbl_train, 2) || (sett.N <= 0))
        str = 'size(tbl_train, 2)';
        disp(' '                                                    );
        disp(['Improper max. number features, set to DEFAULT: ' str]);
        disp('Allowed values <1, size(tbl_train, 2))'               );
        disp('WARNING: values  Z'                                  );

        sett.N = DEFAULT_MAX_NUM_FEAT;
        STR_MAX_NUM_FEAT = num2str(sett.N);
        STR_MAX_NUM_FEAT_STATE = ' (AUTOMATIC)';
    else
        STR_MAX_NUM_FEAT = num2str(sett.N);
        STR_MAX_NUM_FEAT_STATE = '';
    end
end

%  [C] Console output of the input variables and warnings (if occurred)
fprintf('\n'                                                    );
fprintf('INPUT VARIABLES\n'                                     );
fprintf('training matrix:\t\t(OK): specified\n'                 );
fprintf('training labels:\t\t(OK): specified\n'                 );
fprintf('\n'                                                    );
fprintf('OPTIONAL VARIABLES\n'                                  );
fprintf('Cross validations:\t\t(OK): %s%s\n',  ...
    STR_NUM_CROSS_VAL, STR_NUM_CROSS_VAL_STATE                  );
fprintf('K-fold       \t\t\t(OK): %s%s\n',     ...
    STR_K_FOLD, STR_K_FOLD_STATE                                );
fprintf('Classifier:\t\t\t\t(OK): %s%s\n',     ...
    STR_CLASSIFIER, STR_CLASSIFIER_STATE                        );
fprintf('Scoring method:\t\t\t(OK): %s%s\n',   ...
    STR_SCORE, STR_SCORE_STATE                                  );
fprintf('Added features:\t\t\t(OK): %s%s\n',   ...
    STR_ALLOWED_FEAT, STR_ALLOWED_FEAT_STATE                    );
fprintf('Max. num. features:\t\t(OK): %s%s\n', ...
    STR_MAX_NUM_FEAT, STR_MAX_NUM_FEAT_STATE                    );
fprintf('\n'                                                    );

%% Set the variables
out       = struct(); % Empty struct of output variables (features, etc)
var       = struct(); % Empty struct of output variable changes in time
out.S     = [];       % Empty vector of best performing features (idx)
out.SCR   = [];       % Empty vector of regression scores (optional)
out.RMSE  = [];       % Empty vector of root mean squared error (RMSE)
out.RHO   = [];       % Empty vector of Spearman's correlation coefficient
out.P     = [];       % Empty vector of significance level of correlation

total_best_val = inf; % Total best value of SCORE in the SFFS func. call       
iterator = 0;         % Helper variable to iterate over variable changes
loop = 1 + sett.t;    % Control variable (counter of loops in while)

%% Set the possible classes (check classes)
if ((iscell(label_train)) || (length(unique(label_train)) < 2))
    error('Labels must be a numerical vector with >= 2 classes');
else
    ncv = sett.num_cross_validations;
    unq = unique(label_train);
    len = length(label_train);
    cls = sett.classifier;
    
    best_val = [];
    rmse     = [];  
    rho      = [];
    p        = [];
    
    pred_labels = zeros(len, ncv);
    true_labels = zeros(len, ncv);
end

%% Perform Sequential Floating Forward Selection
while (loop > 0)

    % [A] EVALUATE THE UNION OF THE CURRENTLY SELECTED FEATURES WITH EACH 
    %     YET UNSELECTED FEATURE
    bestidx                     = 0;
    best_val(length(out.S) + 1) = inf;   
    rmse(length(out.S) + 1)     = 0;  
    rho(length(out.S) + 1)      = 0;
    p(length(out.S) + 1)        = 0;
    
    for i1 = 1:size(tbl_train, 2)
        if ~ismember(i1, out.S)
            
            % a) Select the best features so far plus a random feature and
            %    set the index vector as the union of S (feat. selected so
            %    far) and the actual iteration number (idx), if feature is
            %    not a member of S yet
            sel = [out.S i1];

            % b) Perform the cross validations (number of validations is
            %    specified by the settings of the algorithm). In each of
            %    the validation steps data is randomly shuffled and next
            %    set (training/testing) data's selected (stratification)
            for cvi = 1:ncv
                     
                % Create temporary vector (predictions/presennted)
                predictions = [];
                presented   = [];

                % ------------------------------------------------------- 
                % Leave-one-Out validation selected
                if (sett.k_fold == 1)

                    % Shuffle the training data befor the validation
                    temp_t  = tbl_train;
                    temp_l  = label_train;
                    rnd_ord = randperm(length(label_train)).';

                    temp_t = temp_t(rnd_ord, :);
                    temp_l = temp_l(rnd_ord, :); 

                    % Iterate over all folds of the cross validation and 
                    % do the training and testing process of the chosen
                    % regression algorithm
                    for fold = 1:length(label_train)

                        t_train = temp_t;
                        l_train = temp_l;
                        t_test  = temp_t(fold, :);

                        t_train(fold, :) = [];
                        l_train(fold, :) = [];

                        % Perform the estimation (regression)   
                        [~, pred_nested] =                             ...
                            perf_regression(t_train(:, sel),           ...
                                l_train, t_test(:, sel), cls);
                        
                        predictions = [predictions; pred_nested(:)];
                        presented   = temp_l;
                    end

                % -------------------------------------------------------  
                % K-fold cross validation selected
                else

                    % Prepare cross validation data and labels for the
                    % cross validation. From this set the k-folds will 
                    % be executed. It also randomizes the data
                    [data, labels, cat_vec, fold_sizes, ~] =           ...
                        get_cv_data(tbl_train, label_train,            ...
                            sett.k_fold, unq);

                    % Iterate over all folds of the cross validation
                    % and do the training and testing process of the
                    % chosen regression algorithm
                    for fold = 1:sett.k_fold

                        % Get the training and testing samples
                        [t_train, l_train, t_test, pres_nested] =      ...
                            get_fold_vectors(data, labels, unq,        ...
                                cat_vec, fold_sizes, fold); 

                        % Perform the estimation (regression)   
                        [~, pred_nested] =                             ...
                            perf_regression(t_train(:, sel),           ...
                                l_train, t_test(:, sel), cls);
                        
                        predictions = [predictions; pred_nested(:)];
                        presented   = [presented;   pres_nested(:)];
                    end
                end

                % Update the true and predicted labels
                pred_labels(:, cvi) = predictions;
                true_labels(:, cvi) = presented;
            end
            
            % c) Compute overall estimation (reg) performance score that 
            %    is achieved so far (default: 'mae' estim. scoring alg.) 
            RMSE_cv = zeros(ncv, 1);
            RHO_cv  = zeros(ncv, 1);
            P_cv    = zeros(ncv, 1);            
            SCR_cv  = zeros(ncv, 1);

            for cvi = 1:ncv
                
                true_l = true_labels(:, cvi);
                pred_l = pred_labels(:, cvi);
                
                % Calculate the regression scores
                RMSE_cv(cvi) =             ...
                    calc_regression_score(pred_l, true_l, 'rmse');
                [RHO_cv(cvi), P_cv(cvi)] = ...
                    corr(pred_l, true_l, 'type', 'Spearman');
                SCR_cv(cvi) =              ...
                    calc_regression_score(pred_l, true_l, sett.score);
            end
            
            % Calculate the mean scores over all cross validations
            RMSE_temp = mean(RMSE_cv);
            RHO_temp  = mean(RHO_cv);
            P_temp    = mean(P_cv);
            SCR_temp  = mean(SCR_cv);
            
            % d) Check whether obtained score is better than the old one 
            %    if yes set the currently included feature as it is best
            %    best candidate and update the best score
            if SCR_temp < best_val(length(out.S) + 1)
                bestidx = i1;
                best_val(length(out.S) + 1) = SCR_temp;
                rmse(length(out.S) + 1)     = RMSE_temp;
                rho(length(out.S) + 1)      = RHO_temp;
                p(length(out.S) + 1)        = P_temp;
                
                iterator                   = +iterator + 1;
                var.num_features(iterator) = length(out.S);
                var.score(iterator)        = SCR_temp;
                var.rmse(iterator)         = RMSE_temp;
                var.rho(iterator)          = RHO_temp;
                var.p(iterator)            = P_temp;
                
                statstr = ['SFFS(' num2str(length(out.S))   ...
                    '): ' num2str(i1) ':'                   ...
                    ' SCORE = ' num2str(SCR_temp)           ...
                    ' RMSE = ' num2str(RMSE_temp)           ...
                    ' RHO = ' num2str(RHO_temp)             ...
                    ' P = ' num2str(P_temp)];
                if (SCR_temp <= total_best_val)
                    statstr = [statstr ' *'];
                end     
                
                disp(statstr);
            end
        end
    end

    % [B] INCLUDE THE BEST FEATURE IN THE SET OF SELECTED FEATURES
    out.S    = [out.S bestidx];
    out.SCR  = [out.SCR best_val(length(out.S))];
    out.RMSE = [out.RMSE rmse(length(out.S))];
    out.RHO  = [out.RHO rho(length(out.S))];
    out.P    = [out.P p(length(out.S))];
    
    % [C] "FLOATING" EXCLUSION
    %     Exclude thus far selected features one at a time and find the 
    %     least useful one. Repeat this step as long as it's evaluation 
    %     score stay better than with the previous best score, by using
    %     feature set of the same size 
    first_exclusion = true;
    while (length(out.S) > 2)
        
        % (1) Find the least useful feature to be excluded from the set
        %     of the best festures selected so far by incrementing over
        %     the set and temporary excluding one feature at the time.
        exclusion_score = inf; 
        excidx = 0;
        for i1 = 1:length(out.S)
            
            % a) Select one feature from the selected feature set S at 
            %    the time, to exclude it from the selected features to  
            %    find the least useful one ("floating" exclusion)
            S_x = setdiff(out.S, out.S(i1));
            
            % b) Perform the cross validations (number of validations is
            %    specified by the settings of the algorithm). In each of
            %    the validation steps data is randomly shuffled and next
            %    set (training/testing) data's selected (stratification)
            for cvi = 1:ncv
                    
                % Create temporary vector (predictions/presennted)
                predictions = [];
                presented   = [];

                % ------------------------------------------------------- 
                % Leave-one-Out validation selected
                if (sett.k_fold == 1)

                    % Shuffle the training data befor the validation
                    temp_t  = tbl_train;
                    temp_l  = label_train;
                    rnd_ord = randperm(length(label_train)).';

                    temp_t = temp_t(rnd_ord, :);
                    temp_l = temp_l(rnd_ord, :); 

                    % Iterate over all folds of the cross validation and 
                    % do the training and testing process of the chosen
                    % regression algorithm
                    for fold = 1:length(label_train)

                        t_train = temp_t;
                        l_train = temp_l;
                        t_test  = temp_t(fold, :);

                        t_train(fold, :) = [];
                        l_train(fold, :) = [];

                        % Perform the estimation (regression)   
                        [~, pred_nested] =                             ...
                            perf_regression(t_train(:, S_x),           ...
                                l_train, t_test(:, S_x), cls);
                        
                        predictions = [predictions; pred_nested(:)];
                        presented   = temp_l;
                    end

                % ------------------------------------------------------- 
                % K-fold cross validation selected
                else

                    % Prepare cross validation data and labels for the
                    % cross validation. From this set the k-folds will 
                    % be executed. It also randomizes the data
                    [data, labels, cat_vec, fold_sizes, ~] =           ...
                        get_cv_data(tbl_train, label_train,            ...
                            sett.k_fold, unq);

                    % Iterate over all folds of the cross validation
                    % and do the training and testing process of the
                    % chosen regression algorithm
                    for fold = 1:sett.k_fold

                        % Get the training and testing samples
                        [t_train, l_train, t_test, pres_nested] =      ...
                            get_fold_vectors(data, labels, unq,        ...
                                cat_vec, fold_sizes, fold); 

                        % Perform the estimation (regression)    
                        [~, pred_nested] =                             ...
                            perf_regression(t_train(:, S_x),           ...
                                l_train, t_test(:, S_x), cls);
                        
                        predictions = [predictions; pred_nested(:)];
                        presented   = [presented;   pres_nested(:)];
                    end
                end

                % Update the true and predicted labels
                pred_labels(:, cvi) = predictions;
                true_labels(:, cvi) = presented;
            end
   
            % c) Compute overall estimation (reg) performance score that 
            %    is achieved so far (default: 'mae' estim. scoring alg.)          
            SCR_cv = zeros(ncv, 1);

            for cvi = 1:ncv
                
                true_l = true_labels(:, cvi);
                pred_l = pred_labels(:, cvi);
                
                % Calculate the regression score
                SCR_cv(cvi) = ...
                    calc_regression_score(pred_l, true_l, sett.score);
            end
            
            % Calculate the mean scores over all cross validations 
            SCR_temp = mean(SCR_cv);   
            
            % d) look for the best feature to exclude from the selected
            %    feature set S. The condition is to find the feature to
            %    produce the lowest regression score. This means that 
            %    after excluding the least useful feature, the algorithm
            %    produces higher precision, than if  other features are 
            %    excluded
            if (SCR_temp < exclusion_score)
                exclusion_score = SCR_temp;
                excidx = out.S(i1);
            end
        end
        
        % (2) If the least useful feature is the most recently added
        %     feature, or no gain can be obtained by exclusion don't
        %     exclude anything and continue with forward selection
        if (first_exclusion && excidx == bestidx) || ...
                (exclusion_score >= best_val(length(out.S) - 1))
            break;
            
        % (3) If exclusion leads to better score, than what has been 
        %     previously obtained with feature set of the same size 
        %     exclude the least useful feature
        else
            idx           = find(out.S == excidx);
            out.S(idx)    = [];
            out.SCR(idx)  = [];
            out.RMSE(idx) = [];
            out.RHO(idx)  = [];
            out.P(idx)    = [];
            
            best_val(length(out.S)) = exclusion_score;
            
            % reset counter after feature exclusion
            loop = 1 + sett.t; 
            
            iterator                   = +iterator + 1;
            var.num_features(iterator) = length(out.S);
            var.score(iterator)        = best_val(length(out.S));
            var.rmse(iterator)         = rmse(length(out.S));
            var.rho(iterator)          = rho(length(out.S));
            var.p(iterator)            = p(length(out.S));
            
            statstr = ['Exclusion (' num2str(length(out.S)) '-1): ' ...
                num2str(excidx) ': ' num2str(exclusion_score)];
            if (exclusion_score <= total_best_val)
                statstr = [statstr ' *'];
            end           
            disp(statstr);
        end
        
        first_exclusion = false;
    end
    
    % [D] STOPPING CONDITIONS AND IMPROVEMENT CHECK
    %     At this phase, this algorithm controls if the length of the
    %     feature set (S) is not larger or equal to specified maximum 
    %     number of features. If the size wasn't specified it was set
    %     to be equal to the number of columns (features) in training
    %     and testing tables. The algorithm also checks, if there was
    %     any improvement obtained in the last iteration.
    
    % (1) Quit if the specified maximum number of features is reached
    if (length(out.S) >= sett.N)
        break;
    end
    
    % (2) Check, if any improvement was made after the last iteration
    if (best_val(length(out.S)) >= total_best_val)
        
        % a) The feature set (S) after inclusion and exclusion didn't 
        %    improve the overall best score => decrease the counter 
        loop = loop - 1; 
    else
        
        % b) An improvement was obtained by the new feature set => do
        %    the reset of the counter to its starting value
        loop = 1 + sett.t; 
        total_best_val = best_val(length(out.S));
    end    
end

%% Back off to the best performing feature set
out.S    = out.S(1:end    - (sett.t + 1) + loop);
out.SCR  = out.SCR(1:end  - (sett.t + 1) + loop);
out.RMSE = out.RMSE(1:end - (sett.t + 1) + loop);
out.RHO  = out.RHO(1:end  - (sett.t + 1) + loop);
out.P    = out.P(1:end    - (sett.t + 1) + loop);

plot_sffs_reg_progress(var);